"""
"""

import time
from typing import NoReturn

import torch
from easydict import EasyDict as ED

try:
    import torch_ecg
except:
    import sys
    from pathlib import Path
    sys.path.append(Path(__file__).absolute().parent.parent)
    import torch_ecg

from torch_ecg.model_configs import (
    # building blocks
    # CNN
    # vgg
    vgg16, vgg16_leadwise,
    # vanilla resnet
    resnet_vanilla_18, resnet_vanilla_34,
    resnet_vanilla_50, resnet_vanilla_101, resnet_vanilla_152,
    resnext_vanilla_50_32x4d, resnext_vanilla_101_32x8d,
    resnet_vanilla_wide_50_2, resnet_vanilla_wide_101_2,
    # custom resnet
    resnet_cpsc2018, resnet_cpsc2018_leadwise,
    # stanford resnet
    resnet_block_stanford, resnet_stanford,
    # cpsc2018 SOTA, legacy
    cpsc_2018, cpsc_2018_leadwise,
    # multi_scopic
    multi_scopic, multi_scopic_leadwise,
    # vanilla dense_net
    dense_net_vanilla,
    # custom dense_net
    dense_net_leadwise,
    # vanilla xception
    xception_vanilla,
    # custom xception
    xception_leadwise,
    # vanilla mobilenets
    mobilenet_v1_vanilla,

    # RNN
    lstm,
    attention,

    # MLP
    linear,

    # ATTN
    non_local,
    squeeze_excitation,
    global_context,


    # downstream tasks
    ATI_CNN_CONFIG,
    ECG_CRNN_CONFIG,
    ECG_SEQ_LAB_NET_CONFIG,
    ECG_SUBTRACT_UNET_CONFIG,
    ECG_UNET_VANILLA_CONFIG,
    ECG_YOLO_CONFIG,
    RR_AF_CRF_CONFIG, RR_AF_VANILLA_CONFIG, RR_LSTM_CONFIG,
)

from torch_ecg.models.cnn.darknet import DarkNet
from torch_ecg.models.cnn.densenet import DenseNet
from torch_ecg.models.cnn.efficientnet import EfficientNet, EfficientNetV2
from torch_ecg.models.cnn.ho_resnet import (
    MidPointResNet, RK4ResNet, RK8ResNet,
)
from torch_ecg.models.cnn.mobilenet import (
    MobileNetV1, MobileNetV2, MobileNetV3,
)
from torch_ecg.models.cnn.multi_scopic import MultiScopicCNN
from torch_ecg.models.cnn.resnet import ResNet
from torch_ecg.models.cnn.vgg import VGG16
from torch_ecg.models.cnn.xception import Xception

from torch_ecg.models.rr_lstm import RR_LSTM
from torch_ecg.models.ecg_crnn import ECG_CRNN
from torch_ecg.models.ecg_seq_lab_net import ECG_SEQ_LAB_NET
from torch_ecg.models.unets.ecg_unet import ECG_UNET
from torch_ecg.models.unets.ecg_subtract_unet import ECG_SUBTRACT_UNET


_DEVICE = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

_IN_CHANNELS = 12
_BATCH_SIZE = 2
_SIG_LEN = 4000
_RR_LEN = 100
_TEST_EXAMPLE = torch.rand((_BATCH_SIZE, _IN_CHANNELS, _SIG_LEN)).to(_DEVICE)
_TEST_RR_EXAMPLE = torch.rand((_RR_LEN, _BATCH_SIZE, 1, )).to(_DEVICE)
_TEST_CLF_CLASSES = ["nsr", "af", "pvc",]
_TEST_DELI_CLASSES = ["qrs", "p", "t",]


@torch.no_grad()
def test_cnn() -> NoReturn:
    """
    """
    start = time.time()
    print("\n"+" Test CNN configs ".center(80,"#")+"\n")
    # test vgg configs
    print("\n"+" Test VGG configs ".center(50,"-")+"\n")
    VGG16.__DEBUG__ = False
    for cfg in ["vgg16", "vgg16_leadwise",]:
        _test_cnn("VGG16", cfg)

    # test resnet configs
    print("\n"+" Test ResNet configs ".center(50,"-")+"\n")
    ResNet.__DEBUG__ = False
    for cfg in [
        "resnet_vanilla_18", "resnet_vanilla_34",
        "resnet_vanilla_50", "resnet_vanilla_101", "resnet_vanilla_152",
        "resnet_vanilla_wide_50_2", "resnet_vanilla_wide_101_2",
        # custom resnet
        "resnet", "resnet_leadwise",
        # TODO: fix bugs in the following
        # "resnext_vanilla_50_32x4d", "resnext_vanilla_101_32x8d",
        # stanford resnet
        # "resnet_stanford",
    ]:
        _test_cnn("ResNet", cfg)

    # test multi_scopic
    print("\n"+" Test MultiScopic configs ".center(50,"-")+"\n")
    MultiScopicCNN.__DEBUG__ = False
    for cfg in ["multi_scopic", "multi_scopic_leadwise",]:
        _test_cnn("MultiScopicCNN", cfg)

    # test densenet
    print("\n"+" Test DenseNet configs ".center(50,"-")+"\n")
    DenseNet.__DEBUG__ = False
    for cfg in ["dense_net_vanilla", "dense_net_leadwise",]:
        _test_cnn("DenseNet", cfg)

    # test xception
    print("\n"+" Test Xception configs ".center(50,"-")+"\n")
    Xception.__DEBUG__ = False
    for cfg in ["xception_vanilla", "xception_leadwise",]:
        _test_cnn("Xception", cfg)

    # test mobilenet
    print("\n"+" Test MobileNet configs ".center(50,"-")+"\n")
    MobileNetV1.__DEBUG__ = False
    MobileNetV2.__DEBUG__ = False
    MobileNetV3.__DEBUG__ = False
    for cfg in ["mobilenet_v1_vanilla",]:
        _test_cnn("MobileNetV1", cfg)

    print(f"total time cost: {time.time()-start:.2f} seconds")
    print("\n"+" Finish testing CNN configs ".center(80,"#")+"\n")

@torch.no_grad()
def _test_cnn(model_name:str, cfg:ED) -> NoReturn:
    """
    """
    try:
        test_model = eval(f"{model_name}(in_channels=_IN_CHANNELS, **{cfg}).to(_DEVICE)")
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"{cfg} output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print(f"{cfg} raises errors\n")
        raise e

@torch.no_grad()
def test_tasks() -> NoReturn:
    """
    """
    start = time.time()
    print("\n"+" Test downstream task configs ".center(80,"#")+"\n")
    # test crnn configs
    print("\n"+" Test ECG_CRNN configs ".center(50,"-")+"\n")
    ECG_CRNN.__DEBUG__ = False
    try:
        test_model = ECG_CRNN(classes=_TEST_CLF_CLASSES, n_leads=_IN_CHANNELS, config=ECG_CRNN_CONFIG).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_CRNN output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_CRNN_CONFIG raises errors\n")
        raise e

    # test seq_lab configs
    print("\n"+" Test ECG_SEQ_LAB_NET configs ".center(50,"-")+"\n")
    ECG_SEQ_LAB_NET.__DEBUG__ = False
    try:
        test_model = ECG_SEQ_LAB_NET(classes=_TEST_DELI_CLASSES, n_leads=_IN_CHANNELS, config=ECG_SEQ_LAB_NET_CONFIG).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_SEQ_LAB_NET output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_SEQ_LAB_NET raises errors\n")
        raise e

    # test unet configs
    print("\n"+" Test ECG_UNET configs ".center(50,"-")+"\n")
    ECG_UNET.__DEBUG__ = False
    try:
        test_model = ECG_UNET(classes=_TEST_DELI_CLASSES, n_leads=_IN_CHANNELS, config=ECG_UNET_VANILLA_CONFIG).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_UNET output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_UNET_VANILLA_CONFIG raises errors\n")
        raise e

    # test subtract_unet configs
    print("\n"+" Test ECG_SUBTRACT_UNET configs ".center(50,"-")+"\n")
    ECG_SUBTRACT_UNET.__DEBUG__ = False
    try:
        test_model = ECG_SUBTRACT_UNET(classes=_TEST_DELI_CLASSES, n_leads=_IN_CHANNELS, config=ECG_SUBTRACT_UNET_CONFIG).to(_DEVICE)
        test_model.eval()
        test_output = test_model(_TEST_EXAMPLE)
        print(f"ECG_SUBTRACT_UNET output shape = {test_output.shape}")
        del test_model
        del test_output
    except Exception as e:
        print("ECG_SUBTRACT_UNET_CONFIG raises errors\n")
        raise e

    # test rr_lstm configs
    print("\n"+" Test RR_LSTM configs ".center(50,"-")+"\n")
    RR_LSTM.__DEBUG__ = False
    for cfg in ["RR_AF_CRF_CONFIG", "RR_AF_VANILLA_CONFIG", "RR_LSTM_CONFIG",]:
        try:
            test_model = eval(f"RR_LSTM(classes=_TEST_CLF_CLASSES, config={cfg}).to(_DEVICE)")
            test_model.eval()
            test_output = test_model(_TEST_RR_EXAMPLE)
            print(f"{cfg} output shape = {test_output.shape}")
            del test_model
            del test_output
        except Exception as e:
            print(f"{cfg} raises errors\n")
            raise e

    print(f"total time cost: {time.time()-start:.2f} seconds")
    print("\n"+" Finish testing downstream task configs ".center(80,"#")+"\n")


def run_test():
    """
    """
    # test CNNs
    test_cnn()
    # test downstream tasks
    test_tasks()


if __name__ == "__main__":
    run_test()



"""
"""

from itertools import repeat
from copy import deepcopy

import numpy as np
from easydict import EasyDict as ED


__all__ = [
    # vanilla xception
    "xception_vanilla",
    # custom xception
    "xception_leadwise",
]


xception_vanilla = ED()
xception_vanilla.fs = 500
xception_vanilla.groups = 1
_base_num_filters = 8
xception_vanilla.entry_flow = ED(
    init_num_filters=[_base_num_filters*4, _base_num_filters*8],
    init_filter_lengths=31,
    init_subsample_lengths=[2,1],
    num_filters=[_base_num_filters*16, _base_num_filters*32, _base_num_filters*91],
    filter_lengths=15,
    subsample_lengths=2,
    subsample_kernels=3,
)
xception_vanilla.middle_flow = ED(
    num_filters=list(repeat(_base_num_filters*91, 8)),
    filter_lengths=13,
)
xception_vanilla.exit_flow = ED(
    final_num_filters=[_base_num_filters*182, _base_num_filters*256],
    final_filter_lengths=3,
    num_filters=[[_base_num_filters*91, _base_num_filters*128]],
    filter_lengths=17,
    subsample_lengths=2,
    subsample_kernels=3,
)

xception_leadwise = ED()
xception_leadwise.fs = 500
xception_leadwise.groups = 12
_base_num_filters = 12 * 2
xception_leadwise.entry_flow = ED(
    init_num_filters=[_base_num_filters*4, _base_num_filters*8],
    init_filter_lengths=31,
    init_subsample_lengths=[2,1],
    num_filters=[_base_num_filters*16, _base_num_filters*32, _base_num_filters*91],
    filter_lengths=15,
    subsample_lengths=2,
    subsample_kernels=3,
)
xception_leadwise.middle_flow = ED(
    num_filters=list(repeat(_base_num_filters*91, 8)),
    filter_lengths=13,
)
xception_leadwise.exit_flow = ED(
    final_num_filters=[_base_num_filters*182, _base_num_filters*256],
    final_filter_lengths=17,
    num_filters=[[_base_num_filters*91, _base_num_filters*128]],
    filter_lengths=3,
    subsample_lengths=2,
    subsample_kernels=3,
)



"""
configs for the basic cnn layers and blocks
"""

from .vgg import (
    vgg_block_basic, vgg_block_mish, vgg_block_swish,
    vgg16, vgg16_leadwise,
)
from .resnet import (
    # building blocks
    resnet_block_basic, resnet_bottle_neck,
    resnet_bottle_neck_B, resnet_bottle_neck_D,
    resnet_block_basic_se, resnet_bottle_neck_se,
    resnet_block_basic_nl, resnet_bottle_neck_nl,
    resnet_block_basic_gc, resnet_bottle_neck_gc,
    # vanilla resnet
    resnet_vanilla_18, resnet_vanilla_34, resnet_vanilla_50,
    resnet_vanilla_101, resnet_vanilla_152,
    resnext_vanilla_50_32x4d, resnext_vanilla_101_32x8d,
    resnet_vanilla_wide_50_2, resnet_vanilla_wide_101_2,
    # cpsc2018 resnet
    resnet_cpsc2018, resnet_cpsc2018_leadwise,
    # smaller resnets
    resnetN, resnetNB, resnetNS, resnetNBS,
    # stanford resnet
    resnet_block_stanford, resnet_stanford,
    # ResNet Nature Communications
    resnet_nature_comm,
    resnet_nature_comm_se, resnet_nature_comm_nl, resnet_nature_comm_gc,
    resnet_nature_comm_bottle_neck, resnet_nature_comm_bottle_neck_se,
    resnet_nature_comm_bottle_neck_gc, resnet_nature_comm_bottle_neck_nl,
    # TResNet
    tresnetF, tresnetP, tresnetN, tresnetS,
    tresnetM, tresnetL, tresnetXL,
    tresnetM_V2,
)
from .cpsc import (
    # cpsc2018 SOTA
    cpsc_block_basic, cpsc_block_mish, cpsc_block_swish,
    cpsc_2018, cpsc_2018_leadwise,
)
from .multi_scopic import (
    multi_scopic_block,
    multi_scopic, multi_scopic_leadwise,
)
from .densenet import (
    # vanilla densenet
    densenet_vanilla,
    # custom densenet
    densenet_leadwise,
)
from .xception import (
    # vanilla xception
    xception_vanilla,
    # custom xception
    xception_leadwise,
)
from .mobilenet import (
    # vanilla mobilenets
    mobilenet_v1_vanilla,
    mobilenet_v2_vanilla,
)


__all__ = [
    # vgg
    "vgg_block_basic", "vgg_block_mish", "vgg_block_swish",
    "vgg16", "vgg16_leadwise",
    # resnet building blocks
    "resnet_block_basic", "resnet_bottle_neck",
    "resnet_bottle_neck_B", "resnet_bottle_neck_D",
    "resnet_block_basic_se", "resnet_bottle_neck_se",
    "resnet_block_basic_nl", "resnet_bottle_neck_nl",
    "resnet_block_basic_gc", "resnet_bottle_neck_gc",
    # smaller resnets
    "resnetN", "resnetNB", "resnetNS", "resnetNBS",
    # vanilla resnet
    "resnet_vanilla_18", "resnet_vanilla_34",
    "resnet_vanilla_50", "resnet_vanilla_101", "resnet_vanilla_152",
    "resnext_vanilla_50_32x4d", "resnext_vanilla_101_32x8d",
    "resnet_vanilla_wide_50_2", "resnet_vanilla_wide_101_2",
    # cpsc2018 resnet
    "resnet_cpsc2018", "resnet_cpsc2018_leadwise",
    # stanford resnet
    "resnet_block_stanford", "resnet_stanford",
    # ResNet Nature Communications
    "resnet_nature_comm",
    "resnet_nature_comm_se", "resnet_nature_comm_nl", "resnet_nature_comm_gc",
    "resnet_nature_comm_bottle_neck", "resnet_nature_comm_bottle_neck_se",
    "resnet_nature_comm_bottle_neck_gc", "resnet_nature_comm_bottle_neck_nl",
    # TresNet
    "tresnetF", "tresnetP", "tresnetN", "tresnetS",
    "tresnetM", "tresnetL", "tresnetXL",
    "tresnetM_V2",
    # cpsc2018 SOTA
    "cpsc_block_basic", "cpsc_block_mish", "cpsc_block_swish",
    "cpsc_2018", "cpsc_2018_leadwise",
    # multi_scopic
    "multi_scopic_block",
    "multi_scopic", "multi_scopic_leadwise",
    # vanilla densenet
    "densenet_vanilla",
    # custom densenet
    "densenet_leadwise",
    # vanilla xception
    "xception_vanilla",
    # custom xception
    "xception_leadwise",
    # vanilla mobilenets
    "mobilenet_v1_vanilla",
    "mobilenet_v2_vanilla",
]



"""
"""

from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Optional, Tuple, NoReturn, Sequence, Any, Set

import numpy as np

from ..cfg import CFG


__all__ = [
    "BaseOutput",
    "ClassificationOutput",
    "MultiLabelClassificationOutput",
    "SequenceTaggingOutput",
    "SequenceLabelingOutput",
    "WaveDelineationOutput",
    "RPeaksDetectionOutput",
]


class BaseOutput(CFG, ABC):
    """
    """
    __name__ = "BaseOutput"

    def __init__(self, *args:Any, **kwargs:Any) -> NoReturn:
        """
        """
        super().__init__(*args, **kwargs)
        pop_fields = [k for k in self if k == "required_fields" or k.startswith("_abc")]
        for f in pop_fields:
            self.pop(f)
        assert all(field in self.keys() for field in self.required_fields()), \
            f"{self.__name__} requires {self.required_fields()}, " \
            f"but `{', '.join(self.required_fields() - set(self.keys()))}` are missing"
        assert all(self[field] is not None for field in self.required_fields()), \
            f"Fields `{', '.join([field for field in self.required_fields() if self[field] is None])}` are not set"

    @abstractmethod
    def required_fields(self) -> Set[str]:
        """
        """
        raise NotImplementedError
    

class ClassificationOutput(BaseOutput):
    """
    """
    __name__ = "ClassificationOutput"

    def __init__(self,
                 *args:Any,
                 classes:Sequence[str]=None,
                 prob:np.ndarray=None,
                 pred:np.ndarray=None,
                 **kwargs:Any) -> NoReturn:
        """
        """
        super().__init__(*args, classes=classes, prob=prob, pred=pred, **kwargs)

    def required_fields(self) -> Set[str]:
        """
        """
        return set(["classes", "prob", "pred",])


class MultiLabelClassificationOutput(BaseOutput):
    """
    """
    __name__ = "MultiLabelClassificationOutput"

    def __init__(self,
                 *args:Any,
                 classes:Sequence[str]=None,
                 thr:float=None,
                 prob:np.ndarray=None,
                 pred:np.ndarray=None,
                 **kwargs:Any) -> NoReturn:
        """
        """
        super().__init__(*args, classes=classes, thr=thr, prob=prob, pred=pred, **kwargs)

    def required_fields(self) -> Set[str]:
        """
        """
        return set(["classes", "thr", "prob", "pred",])


class SequenceTaggingOutput(BaseOutput):
    """
    """
    __name__ = "SequenceTaggingOutput"

    def __init__(self,
                 *args:Any,
                 classes:Sequence[str]=None,
                 prob:np.ndarray=None,
                 pred:np.ndarray=None,
                 **kwargs:Any) -> NoReturn:
        """
        """
        super().__init__(*args, classes=classes, prob=prob, pred=pred, **kwargs)

    def required_fields(self) -> Set[str]:
        """
        """
        return set(["classes", "prob", "pred",])

# alias
SequenceLabelingOutput = SequenceTaggingOutput
SequenceLabelingOutput.__name__ = "SequenceLabelingOutput"


class WaveDelineationOutput(SequenceTaggingOutput):
    """
    """
    __name__ = "WaveDelineationOutput"

    def __init__(self,
                 *args:Any,
                 classes:Sequence[str]=None,
                 prob:np.ndarray=None,
                 mask:np.ndarray=None,
                 **kwargs:Any) -> NoReturn:
        """
        """
        super().__init__(*args, classes=classes, prob=prob, pred=mask, mask=mask, **kwargs)
        self.pop("pred")

    def required_fields(self) -> Set[str]:
        """
        """
        return set(["classes", "prob", "mask",])


class RPeaksDetectionOutput(BaseOutput):
    """
    """
    __name__ = "RPeaksDetectionOutput"

    def __init__(self,
                 *args:Any,
                 rpeak_indices:Sequence[Sequence[int]]=None,
                 prob:np.ndarray=None,
                 pred:np.ndarray=None,
                 **kwargs:Any) -> NoReturn:
        """
        """
        super().__init__(*args, rpeak_indices=rpeak_indices, prob=prob, **kwargs)

    def required_fields(self) -> Set[str]:
        """
        """
        return set(["rpeak_indices", "prob",])



"""
"""

from typing import NoReturn, Optional, Any
from numbers import Real
import warnings

import torch

from .._preprocessors.base import (
    preprocess_multi_lead_signal,
)


__all__ = ["BandPass",]


class BandPass(torch.nn.Module):
    """
    """
    __name__ = "BandPass"

    def __init__(self,
                 fs:Real,
                 lowcut:Optional[Real]=0.5,
                 highcut:Optional[Real]=45,
                 inplace:bool=True,
                 **kwargs:Any) -> NoReturn:
        """ finished, checked,

        Parameters
        ----------
        fs: real number,
            sampling frequency of the ECG signal to be filtered
        lowcut: real number, optional,
            low cutoff frequency
        highcut: real number, optional,
            high cutoff frequency
        inplace: bool, default True,
            if True, the preprocessor will modify the input signal
        kwargs: keyword arguments,
        """
        super().__init__()
        self.fs = fs
        self.lowcut = lowcut
        self.highcut = highcut
        assert any([self.lowcut is not None, self.highcut is not None]), \
            "At least one of lowcut and highcut should be set"
        if not self.lowcut:
            self.lowcut = 0
        if not self.highcut:
            self.highcut = float("inf")
        self.inplace = inplace

    def forward(self, sig:torch.Tensor) -> torch.Tensor:
        """ finished, checked,

        apply the preprocessor to `sig`

        Parameters
        ----------
        sig: Tensor,
            the ECG signals, of shape (batch, lead, siglen)

        Returns
        -------
        filtered_sig: Tensor,
            the bandpass filtered ECG signals, of shape (batch, lead, siglen)
        """
        if not self.inplace:
            sig = sig.clone()
        sig = torch.as_tensor(
            preprocess_multi_lead_signal(
                raw_sig=sig.cpu().numpy(),
                fs=self.fs,
                band_fs=[self.lowcut, self.highcut],
            ).copy(),
            dtype=sig.dtype,
            device=sig.device,
        )
        return sig



"""
"""

from .cnn import (
    # vgg
    vgg_block_basic, vgg_block_mish, vgg_block_swish,
    vgg16, vgg16_leadwise,
    # ResNet
    # vanilla resnet
    resnet_block_basic, resnet_bottle_neck,
    resnet_bottle_neck_B, resnet_bottle_neck_D,
    resnet_block_basic_se, resnet_bottle_neck_se,
    resnet_block_basic_nl, resnet_bottle_neck_nl,
    resnet_block_basic_gc, resnet_bottle_neck_gc,
    # vanilla resnet
    resnet_vanilla_18, resnet_vanilla_34, resnet_vanilla_50,
    resnet_vanilla_101, resnet_vanilla_152,
    resnext_vanilla_50_32x4d, resnext_vanilla_101_32x8d,
    resnet_vanilla_wide_50_2, resnet_vanilla_wide_101_2,
    # cpsc2018 resnet
    resnet_cpsc2018, resnet_cpsc2018_leadwise,
    # smaller resnets
    resnetN, resnetNB, resnetNS, resnetNBS,
    # stanford resnet
    resnet_block_stanford, resnet_stanford,
    # ResNet Nature Communications
    resnet_nature_comm,
    resnet_nature_comm_se, resnet_nature_comm_nl, resnet_nature_comm_gc,
    resnet_nature_comm_bottle_neck, resnet_nature_comm_bottle_neck_se,
    resnet_nature_comm_bottle_neck_gc, resnet_nature_comm_bottle_neck_nl,
    # TResNet
    tresnetF, tresnetP, tresnetN, tresnetS,
    tresnetM, tresnetL, tresnetXL,
    tresnetM_V2,
    # cpsc2018 SOTA
    cpsc_block_basic, cpsc_block_mish, cpsc_block_swish,
    cpsc_2018, cpsc_2018_leadwise,
    # multi_scopic
    multi_scopic_block,
    multi_scopic, multi_scopic_leadwise,
    # vanilla densenet
    densenet_vanilla,
    # custom densenet
    densenet_leadwise,
    # vanilla xception
    xception_vanilla,
    # custom xception
    xception_leadwise,
    # vanilla mobilenets
    mobilenet_v1_vanilla,
)
from .rnn import lstm, attention
from .mlp import linear
from .attn import (
    non_local, squeeze_excitation, global_context,
    transformer,
)

from .ati_cnn import ATI_CNN_CONFIG
from .ecg_crnn import ECG_CRNN_CONFIG
from .ecg_seq_lab_net import ECG_SEQ_LAB_NET_CONFIG
from .ecg_subtract_unet import ECG_SUBTRACT_UNET_CONFIG
from .ecg_unet import ECG_UNET_VANILLA_CONFIG
from .ecg_yolo import ECG_YOLO_CONFIG
from .rr_lstm import RR_AF_CRF_CONFIG, RR_AF_VANILLA_CONFIG, RR_LSTM_CONFIG


__all__ = [
    # building blocks

    # CNN
    # vgg
    "vgg_block_basic", "vgg_block_mish", "vgg_block_swish",
    "vgg16", "vgg16_leadwise",
    # ResNet
    # resnet building blocks
    "resnet_block_basic", "resnet_bottle_neck",
    "resnet_bottle_neck_B", "resnet_bottle_neck_D",
    "resnet_block_basic_se", "resnet_bottle_neck_se",
    "resnet_block_basic_nl", "resnet_bottle_neck_nl",
    "resnet_block_basic_gc", "resnet_bottle_neck_gc",
    # smaller resnets
    "resnetN", "resnetNB", "resnetNS", "resnetNBS",
    # vanilla resnet
    "resnet_vanilla_18", "resnet_vanilla_34",
    "resnet_vanilla_50", "resnet_vanilla_101", "resnet_vanilla_152",
    "resnext_vanilla_50_32x4d", "resnext_vanilla_101_32x8d",
    "resnet_vanilla_wide_50_2", "resnet_vanilla_wide_101_2",
    # cpsc2018 resnet
    "resnet_cpsc2018", "resnet_cpsc2018_leadwise",
    # stanford resnet
    "resnet_block_stanford", "resnet_stanford",
    # ResNet Nature Communications
    "resnet_nature_comm",
    "resnet_nature_comm_se", "resnet_nature_comm_nl", "resnet_nature_comm_gc",
    "resnet_nature_comm_bottle_neck", "resnet_nature_comm_bottle_neck_se",
    "resnet_nature_comm_bottle_neck_gc", "resnet_nature_comm_bottle_neck_nl",
    # TresNet
    "tresnetF", "tresnetP", "tresnetN", "tresnetS",
    "tresnetM", "tresnetL", "tresnetXL",
    "tresnetM_V2",
    # cpsc2018 SOTA
    "cpsc_block_basic", "cpsc_block_mish", "cpsc_block_swish",
    "cpsc_2018", "cpsc_2018_leadwise",
    # multi_scopic
    "multi_scopic_block",
    "multi_scopic", "multi_scopic_leadwise",
    # vanilla densenet
    "densenet_vanilla",
    # custom densenet
    "densenet_leadwise",
    # vanilla xception
    "xception_vanilla",
    # custom xception
    "xception_leadwise",
    # vanilla mobilenets
    "mobilenet_v1_vanilla",

    # RNN
    "lstm",
    "attention",

    # MLP
    "linear",

    # ATTN
    "non_local",
    "squeeze_excitation",
    "global_context",
    "transformer",


    # downstream tasks
    "ATI_CNN_CONFIG",
    "ECG_CRNN_CONFIG",
    "ECG_SEQ_LAB_NET_CONFIG",
    "ECG_SUBTRACT_UNET_CONFIG",
    "ECG_UNET_VANILLA_CONFIG",
    "ECG_YOLO_CONFIG",
    "RR_AF_CRF_CONFIG", "RR_AF_VANILLA_CONFIG", "RR_LSTM_CONFIG",
]



# -*- coding: utf-8 -*-
"""
"""

import io
from pathlib import Path
from datetime import datetime
from typing import Union, Optional, Any, List, Dict, Tuple, NoReturn
from numbers import Real

import numpy as np
np.set_printoptions(precision=5, suppress=True)
import pandas as pd
from scipy.io import loadmat

from ..aux_data.cinc2020_aux_data import (
    dx_mapping_all, dx_mapping_scored, dx_mapping_unscored,
    normalize_class, abbr_to_snomed_ct_code,
)
from ..base import CPSCDataBase, DEFAULT_FIG_SIZE_PER_SEC


__all__ = [
    "CPSC2018",
    "compute_metrics",
]


class CPSC2018(CPSCDataBase):
    """

    The China Physiological Signal Challenge 2018:
    Automatic identification of the rhythm/morphology abnormalities in 12-lead ECGs

    ABOUT CPSC2018
    --------------
    1. training set contains 6,877 (female: 3178; male: 3699) 12 leads ECG recordings lasting from 6 s to just 60 s
    2. ECG recordings were sampled as 500 Hz
    3. the training data can be downloaded using links in Ref.[1], but the link in Ref.[2] is recommended. File structure will be assumed to follow Ref.[2]
    4. the training data are in the `channel first` format
    5. types of abnormal rhythm/morphology + normal in the training set:
            name                                    abbr.       number of records
        (0) Normal                                  N           918
        (1) Atrial fibrillation                     AF          1098
        (2) First-degree atrioventricular block     I-AVB       704
        (3) Left bundle brunch block                LBBB        207
        (4) Right bundle brunch block               RBBB        1695
        (5) Premature atrial contraction            PAC         556
        (6) Premature ventricular contraction       PVC         672
        (7) ST-segment depression                   STD         825
        (8) ST-segment elevated                     STE         202
    6. ordering of the leads in the data of all the records are
        ["I", "II", "III", "aVR", "aVL", "aVF", "V1", "V2", "V3", "V4", "V5", "V6"]
    7. meanings in the .hea files: to write
    8. knowledge about the abnormal rhythms: ref. cls.get_disease_knowledge

    Update
    ------
    CINC2020 (ref. [2]) released totally 3453 unused training data of CPSC2018, whose filenames start with "Q".
    These file names are not "continuous". The last record is "Q3581"

    NOTE
    ----
    1. Ages of records A0608, A1549, A1876, A2299, A5990 are "NaN"

    ISSUES
    ------

    Usage
    -----
    1. ecg arrythmia detection

    References
    ----------
    [1] http://2018.icbeb.org/#
    [2] https://physionetchallenges.github.io/2020/
    """

    def __init__(self,
                 db_dir:Union[str,Path],
                 working_dir:Optional[Union[str,Path]]=None,
                 verbose:int=2,
                 **kwargs:Any) -> NoReturn:
        """ finished, to be improved,

        Parameters
        ----------
        db_dir: str or Path,
            storage path of the database
        working_dir: str or Path, optional,
            working directory, to store intermediate files and log file
        verbose: int, default 2,
            log verbosity
        kwargs: auxilliary key word arguments
        """
        super().__init__(db_name="cpsc2018", db_dir=db_dir, working_dir=working_dir, verbose=verbose, **kwargs)

        self.fs = 500
        self.spacing = 1000 / self.fs
        self.rec_ext = "mat"
        self.ann_ext = "hea"
        self._all_records = None
        self._ls_rec()

        self.nb_records = 6877
        self.all_leads = ["I", "II", "III", "aVR", "aVL", "aVF", "V1", "V2", "V3", "V4", "V5", "V6",]
        self.all_diagnosis = ["N", "AF", "I-AVB", "LBBB", "RBBB", "PAC", "PVC", "STD", "STE",]
        self.all_diagnosis_original = sorted(["Normal", "AF", "I-AVB", "LBBB", "RBBB", "PAC", "PVC", "STD", "STE",])
        self.diagnosis_abbr_to_full = {
            "N": "Normal",
            "AF": "Atrial fibrillation",
            "I-AVB": "First-degree atrioventricular block",
            "LBBB": "Left bundle brunch block",
            "RBBB": "Right bundle brunch block",
            "PAC": "Premature atrial contraction",
            "PVC": "Premature ventricular contraction",
            "STD": "ST-segment depression",
            "STE": "ST-segment elevated",
        }

        self.ann_items = [
            "rec_name",
            "nb_leads",
            "fs",
            "nb_samples",
            "datetime",
            "age",
            "sex",
            "diagnosis",
            "medical_prescription",
            "history",
            "symptom_or_surgery",
            "df_leads",
        ]

    def _ls_rec(self) -> NoReturn:
        """
        """
        self._all_records = [
            item.with_suffix("").name \
                for item in self.db_dir.glob(f"*.{self.rec_ext}")
        ]

    def get_subject_id(self, rec_no:Union[int,str]) -> int:
        """ not finished,

        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record

        Returns
        -------
        pid: int,
            the `subject_id` corr. to `rec_no`
        """
        raise NotImplementedError

    def load_data(self, rec_no:Union[int,str], data_format="channel_first", units:str="mV",) -> np.ndarray:
        """ finished, checked,

        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record,
            int only supported for the original CPSC2018 dataset
        data_format: str, default "channel_first",
            format of the ecg data, "channels_last" or "channels_first" (original)
        units: str, default "mV",
            units of the output signal, can also be "μV", with an alias of "uV"
        
        Returns
        -------
        data: ndarray,
            the ecg data
        """
        if isinstance(rec_no, int):
            assert rec_no in range(1, self.nb_records+1), f"rec_no should be in range(1,{self.nb_records+1})"
            rec_no = f"A{rec_no:04d}"
        rec_fp = self.db_dir / f"{rec_no}.{self.rec_ext}"
        data = loadmat(str(rec_fp))
        data = np.asarray(data["val"], dtype=np.float64)
        if data_format == "channels_last":
            data = data.T

        if units.lower() == "mv" and self._auto_infer_units(data) != "mV":
            data /= 1000
        elif units.lower() in ["uv", "μv",] and self._auto_infer_units(data) != "μV":
            data *= 1000
        
        return data

    def load_ann(self, rec_no:Union[int,str], keep_original:bool=True) -> dict:
        """ finished, checked,
        
        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record,
            int only supported for the original CPSC2018 dataset
        keep_original: bool, default True,
            keep the original annotations or not,
            mainly concerning "N" and "Normal" ("SNR" for the newer version)
        
        Returns
        -------
        ann_dict, dict,
            the annotations with items: ref. self.ann_items
        """
        if isinstance(rec_no, int):
            assert rec_no in range(1, self.nb_records+1), f"rec_no should be in range(1, {self.nb_records+1})"
            rec_no = f"A{rec_no:04d}"
        ann_fp = self.db_dir / f"{rec_no}.{self.ann_ext}"
        header_data = ann_fp.read_text().splitlines()

        ann_dict = {}
        ann_dict["rec_name"], ann_dict["nb_leads"], ann_dict["fs"], ann_dict["nb_samples"], ann_dict["datetime"], daytime \
            = header_data[0].split(" ")

        ann_dict["nb_leads"] = int(ann_dict["nb_leads"])
        ann_dict["fs"] = int(ann_dict["fs"])
        ann_dict["nb_samples"] = int(ann_dict["nb_samples"])
        ann_dict["datetime"] = datetime.strptime(" ".join([ann_dict["datetime"], daytime]), "%d-%b-%Y %H:%M:%S")
        try: # see NOTE. 1.
            ann_dict["age"] = int([l for l in header_data if l.startswith("#Age")][0].split(": ")[-1])
        except:
            ann_dict["age"] = np.nan
        try:
            ann_dict["sex"] = [l for l in header_data if l.startswith("#Sex")][0].split(": ")[-1]
        except:
            ann_dict["sex"] = "Unknown"
        try:
            ann_dict["medical_prescription"] = [l for l in header_data if l.startswith("#Rx")][0].split(": ")[-1]
        except:
            ann_dict["medical_prescription"] = "Unknown"
        try:
            ann_dict["history"] = [l for l in header_data if l.startswith("#Hx")][0].split(": ")[-1]
        except:
            ann_dict["history"] = "Unknown"
        try:
            ann_dict["symptom_or_surgery"] = [l for l in header_data if l.startswith("#Sx")][0].split(": ")[-1]
        except:
            ann_dict["symptom_or_surgery"] = "Unknown"

        l_Dx = [l for l in header_data if l.startswith("#Dx")][0].split(": ")[-1].split(",")
        ann_dict["diagnosis"], ann_dict["diagnosis_scored"] = self._parse_diagnosis(l_Dx)

        ann_dict["df_leads"] = self._parse_leads(header_data[1:13])

        return ann_dict

    def _parse_diagnosis(self, l_Dx:List[str]) -> Tuple[dict, dict]:
        """ finished, checked,

        Parameters
        ----------
        l_Dx: list of str,
            raw information of diagnosis, read from a header file

        Returns
        -------
        diag_dict:, dict,
            diagnosis, including SNOMED CT Codes, fullnames and abbreviations of each diagnosis
        diag_scored_dict: dict,
            the scored items in `diag_dict`
        """
        diag_dict, diag_scored_dict = {}, {}
        try:
            diag_dict["diagnosis_code"] = [item for item in l_Dx]
            # selection = dx_mapping_all["SNOMED CT Code"].isin(diag_dict["diagnosis_code"])
            # diag_dict["diagnosis_abbr"] = dx_mapping_all[selection]["Abbreviation"].tolist()
            # diag_dict["diagnosis_fullname"] = dx_mapping_all[selection]["Dx"].tolist()
            diag_dict["diagnosis_abbr"] = \
                [ dx_mapping_all[dx_mapping_all["SNOMED CT Code"]==dc]["Abbreviation"].values[0] \
                    for dc in diag_dict["diagnosis_code"] ]
            diag_dict["diagnosis_fullname"] = \
                [ dx_mapping_all[dx_mapping_all["SNOMED CT Code"]==dc]["Dx"].values[0] \
                    for dc in diag_dict["diagnosis_code"] ]
            scored_indices = np.isin(diag_dict["diagnosis_code"], dx_mapping_scored["SNOMED CT Code"].values)
            diag_scored_dict["diagnosis_code"] = \
                [ item for idx, item in enumerate(diag_dict["diagnosis_code"]) \
                    if scored_indices[idx] ]
            diag_scored_dict["diagnosis_abbr"] = \
                [ item for idx, item in enumerate(diag_dict["diagnosis_abbr"]) \
                    if scored_indices[idx] ]
            diag_scored_dict["diagnosis_fullname"] = \
                [ item for idx, item in enumerate(diag_dict["diagnosis_fullname"]) \
                    if scored_indices[idx] ]
        except:  # the old version, the Dx"s are abbreviations
            diag_dict["diagnosis_abbr"] = diag_dict["diagnosis_code"]
            selection = dx_mapping_all["Abbreviation"].isin(diag_dict["diagnosis_abbr"])
            diag_dict["diagnosis_fullname"] = dx_mapping_all[selection]["Dx"].tolist()
        # if not keep_original:
        #     for idx, d in enumerate(ann_dict["diagnosis_abbr"]):
        #         if d in ["Normal", "NSR"]:
        #             ann_dict["diagnosis_abbr"] = ["N"]
        return diag_dict, diag_scored_dict

    def _parse_leads(self, l_leads_data:List[str]) -> pd.DataFrame:
        """ finished, checked,

        Parameters
        ----------
        l_leads_data: list of str,
            raw information of each lead, read from a header file

        Returns
        -------
        df_leads: DataFrame,
            infomation of each leads in the format of DataFrame
        """
        df_leads = pd.read_csv(io.StringIO("\n".join(l_leads_data)), delim_whitespace=True, header=None)
        df_leads.columns = [
            "filename", "fmt+byte_offset", "adc_gain+units", "adc_res", "adc_zero",
            "init_value", "checksum", "block_size", "lead_name",
        ]
        df_leads["fmt"] = df_leads["fmt+byte_offset"].apply(lambda s: s.split("+")[0])
        df_leads["byte_offset"] = df_leads["fmt+byte_offset"].apply(lambda s: s.split("+")[1])
        df_leads["adc_gain"] = df_leads["adc_gain+units"].apply(lambda s: s.split("/")[0])
        df_leads["adc_units"] = df_leads["adc_gain+units"].apply(lambda s: s.split("/")[1])
        for k in ["byte_offset", "adc_gain", "adc_res", "adc_zero", "init_value", "checksum",]:
            df_leads[k] = df_leads[k].apply(lambda s: int(s))
        df_leads["baseline"] = df_leads["adc_zero"]
        df_leads = df_leads[[
            "filename", "fmt", "byte_offset", "adc_gain", "adc_units", "adc_res", "adc_zero", "baseline",
            "init_value", "checksum", "block_size", "lead_name",
        ]]
        df_leads.index = df_leads["lead_name"]
        df_leads.index.name = None
        return df_leads

    def get_labels(self, rec_no:Union[int,str], keep_original:bool=False) -> List[str]:
        """ finished, checked,
        
        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record,
            int only supported for the original CPSC2018 dataset
        keep_original: bool, default False,
            keep the original annotations or not,
            mainly concerning "N" and "Normal"
        
        Returns
        -------
        labels, list,
            the list of labels (abbr. diagnosis)
        """
        ann_dict = self.load_ann(rec_no, keep_original=keep_original)
        labels = ann_dict["diagnosis"]
        return labels

    def get_diagnosis(self, rec_no:Union[int,str], full_name:bool=True) -> List[str]:
        """ finished, checked,
        
        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record,
            int only supported for the original CPSC2018 dataset
        full_name: bool, default True,
            full name of the diagnosis or short name of it (ref. self.diagnosis_abbr_to_full)
        
        Returns
        -------
        diagonosis, list,
            the list of (full) diagnosis
        """
        diagonosis = self.get_labels(rec_no)
        if full_name:
            diagonosis = diagonosis["diagnosis_fullname"]
        else:
            diagonosis = diagonosis["diagnosis_abbr"]
        return diagonosis

    def get_subject_info(self, rec_no:Union[int,str], items:Optional[List[str]]=None) -> dict:
        """ finished, checked,

        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record,
            int only supported for the original CPSC2018 dataset
        items: list of str, optional,
            items of the subject information (e.g. sex, age, etc.)
        
        Returns
        -------
        subject_info, dict,
        """
        if items is None or len(items) == 0:
            info_items = [
                "age", "sex", "medical_prescription", "history", "symptom_or_surgery",
            ]
        else:
            info_items = items
        ann_dict = self.load_ann(rec_no)
        subject_info = {item: ann_dict[item] for item in info_items}

        return subject_info

    def save_challenge_predictions(self,
                                   rec_no:Union[int,str],
                                   output_dir:Union[str,Path],
                                   scores:List[Real],
                                   labels:List[int],
                                   classes:List[str],) -> NoReturn:
        """ finished, checked,

        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record,
            int only supported for the original CPSC2018 dataset
        output_dir: str or Path,
            directory to save the predictions
        scores: list of real,
            ...
        labels: list of int,
            0 or 1
        classes: list of str,
            ...
        """
        if isinstance(rec_no, str):
            rec_no = int(rec_no[1:])
        assert rec_no in range(1, self.nb_records+1), f"rec_no should be in range(1, {self.nb_records+1})"
        recording = self._all_records[rec_no]
        new_file = recording + ".csv"
        output_file = Path(output_dir) / new_file

        # Include the filename as the recording number
        recording_string = f"#{recording}"
        class_string = ",".join(classes)
        label_string = ",".join(str(i) for i in labels)
        score_string = ",".join(str(i) for i in scores)

        output_file.write_text("\n".join([recording_string, class_string, label_string, score_string, ""]))

    def plot(self, rec_no:Union[int,str], ticks_granularity:int=0, leads:Optional[Union[str, List[str]]]=None, **kwargs:Any) -> NoReturn:
        """ finished, checked,

        Parameters
        ----------
        rec_no: int or str,
            number of the record, NOTE that rec_no starts from 1; or name of the record,
            int only supported for the original CPSC2018 dataset
        ticks_granularity: int, default 0,
            the granularity to plot axis ticks, the higher the more,
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads: str or list of str, optional,
            the leads to plot
        kwargs: auxilliary key word arguments
        """
        if isinstance(rec_no, str):
            rec_no = int(rec_no[1:])
        assert rec_no in range(1, self.nb_records+1), f"rec_no should be in range(1, {self.nb_records+1})"
        if "plt" not in dir():
            import matplotlib.pyplot as plt
        if leads is None or leads == "all":
            leads = self.all_leads
        assert all([l in self.all_leads for l in leads])

        lead_list = self.load_ann(rec_no)["df_leads"]["lead_name"].tolist()
        lead_indices = [lead_list.index(l) for l in leads]
        data = self.load_data(rec_no, data_format="channel_first", units="μV")[lead_indices]
        y_ranges = np.max(np.abs(data), axis=1) + 100

        diag = self.get_diagnosis(rec_no, full_name=False)

        nb_leads = len(leads)

        t = np.arange(data.shape[1]) / self.fs
        duration = len(t) / self.fs
        fig_sz_w = int(round(DEFAULT_FIG_SIZE_PER_SEC * duration))
        fig_sz_h = 6 * y_ranges / 1500
        fig, axes = plt.subplots(nb_leads, 1, sharex=True, figsize=(fig_sz_w, np.sum(fig_sz_h)))
        for idx in range(nb_leads):
            axes[idx].plot(t, data[idx], color="black", linewidth="2.0", label=f"lead - {leads[idx]}")
            axes[idx].axhline(y=0, linestyle="-", linewidth="1.0", color="red")
            if ticks_granularity >= 1:
                axes[idx].xaxis.set_major_locator(plt.MultipleLocator(0.2))
                axes[idx].yaxis.set_major_locator(plt.MultipleLocator(500))
                axes[idx].grid(which="major", linestyle="-", linewidth="0.5", color="red")
            if ticks_granularity >= 2:
                axes[idx].xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                axes[idx].yaxis.set_minor_locator(plt.MultipleLocator(100))
                axes[idx].grid(which="minor", linestyle=":", linewidth="0.5", color="black")
            axes[idx].plot([], [], " ", label=f"labels - {','.join(diag)}")
            axes[idx].legend(loc="upper left")
            axes[idx].set_xlim(t[0], t[-1])
            axes[idx].set_ylim(-y_ranges[idx], y_ranges[idx])
            axes[idx].set_xlabel("Time [s]")
            axes[idx].set_ylabel("Voltage [μV]")
        plt.subplots_adjust(hspace=0.2)
        if kwargs.get("save_path", None):
            plt.savefig(kwargs["save_path"], dpi=200, bbox_inches="tight")
        else:
            plt.show()

    @property
    def url(self) -> List[str]:
        return [
            "http://hhbucket.oss-cn-hongkong.aliyuncs.com/TrainingSet1.zip",
            "http://hhbucket.oss-cn-hongkong.aliyuncs.com/TrainingSet2.zip",
            "http://hhbucket.oss-cn-hongkong.aliyuncs.com/TrainingSet3.zip",
        ]


def compute_metrics():
    """
    """
    raise NotImplementedError



"""
"""

import torch

try:
    import torch_ecg
except:
    import sys
    from pathlib import Path
    sys.path.append(Path(__file__).absolute().parent.parent)
    import torch_ecg

from torch_ecg.models.loss import (
    BCEWithLogitsWithClassWeightLoss,
    MaskedBCEWithLogitsLoss,
    FocalLoss,
    AsymmetricLoss,
)


inp = torch.tensor([[10.0,-10.0],[-10.0,10.0]])
targ_1 = torch.tensor([[1.0,0.0],[0.0,1.0]])
targ_0 = torch.tensor([[0.0,1.0],[1.0,0.0]])
targ_1_soft = torch.tensor([[0.9,0.1],[0.1,0.9]])
targ_0_soft = torch.tensor([[0.1,0.9],[0.9,0.1]])

criterion_bce_cw = BCEWithLogitsWithClassWeightLoss(class_weight=torch.tensor([[1.0,2.0]]))
criterion_focal = FocalLoss(class_weight=torch.tensor([[1.0,2.0]]), multi_label=True)
criterion_asl = AsymmetricLoss()

criterion_mbce = MaskedBCEWithLogitsLoss()



"""
data generator for feeding data into pytorch models
"""

import json, time, textwrap
from pathlib import Path
import multiprocessing as mp
from random import shuffle, randint, sample
from copy import deepcopy
from typing import Union, Optional, List, Tuple, Dict, Sequence, Set, NoReturn

import numpy as np
np.set_printoptions(precision=5, suppress=True)
try:
    from tqdm.auto import tqdm
except ModuleNotFoundError:
    from tqdm import tqdm
import torch
from torch.utils.data.dataset import Dataset

from ....cfg import CFG
from ....databases import CINC2021 as CR
from ....utils.misc import ensure_siglen, dict_to_str, list_sum, ReprMixin
from ....utils.utils_signal import normalize, remove_spikes_naive
from ....utils.ecg_arrhythmia_knowledge import Standard12Leads
from ...._preprocessors import PreprocManager

from .cinc2021_cfg import CINC2021TrainCfg


__all__ = [
    "CINC2021Dataset",
]


class CINC2021Dataset(ReprMixin, Dataset):
    """
    """
    __DEBUG__ = False
    __name__ = "CINC2021Dataset"

    def __init__(self, config:Optional[CFG]=None, training:bool=True, lazy:bool=True) -> NoReturn:
        """ finished, checked,

        Parameters
        ----------
        config: dict,
            configurations for the Dataset,
            ref. `CINC2021TrainCfg`,
            a simple example is:
            >>> config = deepcopy(CINC2021TrainCfg)
            >>> config.db_dir = "some/path/to/db"
            >>> dataset = CINC2021Dataset(config, training=True, lazy=False)
        training: bool, default True,
            if True, the training set will be loaded, otherwise the test set
        lazy: bool, default True,
            if True, the data will not be loaded immediately,
        """
        super().__init__()
        self.config = deepcopy(config)
        assert self.config.db_dir is not None, "db_dir must be specified"
        self.config.db_dir = Path(self.config.db_dir)
        self._TRANCHES = self.config.tranche_classes.keys()  # ["A", "B", "AB", "E", "F", "G",]
        self.reader = CR(db_dir=self.config.db_dir)
        self.tranches = self.config.tranches_for_training
        self.training = training
        if self.config.torch_dtype == torch.float64:
            self.dtype = np.float64
        else:
            self.dtype = np.float32
        assert not self.tranches or self.tranches in self._TRANCHES
        if self.tranches:
            self.all_classes = self.config.tranche_classes[self.tranches]
            self.class_weights = self.config.tranche_class_weights[self.tranches]
        else:
            self.all_classes = self.config.classes
            self.class_weights = self.config.class_weights
        self.config.all_classes = deepcopy(self.all_classes)
        self.n_classes = len(self.all_classes)
        # print(f"tranches = {self.tranches}, all_classes = {self.all_classes}")
        # print(f"class_weights = {dict_to_str(self.class_weights)}")
        cw = np.zeros((len(self.class_weights),), dtype=self.dtype)
        for idx, c in enumerate(self.all_classes):
            cw[idx] = self.class_weights[c]
        self.class_weights = torch.from_numpy(cw.astype(self.dtype)).view(1, self.n_classes)
        # validation also goes in batches, hence length has to be fixed
        self.siglen = self.config.input_len
        self.lazy = lazy

        self._indices = [Standard12Leads.index(l) for l in self.config.leads]

        self.records = self._train_test_split(self.config.train_ratio, force_recompute=False)
        # TODO: consider using `remove_spikes_naive` to treat these exceptional records
        self.records = [r for r in self.records if r not in self.reader.exceptional_records\
            and (self.reader.get_data_filepath(r)).is_file()]
        if self.__DEBUG__:
            self.records = sample(self.records, int(len(self.records) * 0.01))

        ppm_config = CFG(random=False)
        ppm_config.update(self.config)
        self.ppm = PreprocManager.from_config(ppm_config)
        self.ppm.rearrange(["bandpass", "normalize"])

        self._signals = np.array([], dtype=self.dtype).reshape(0, len(self.config.leads), self.siglen)
        self._labels = np.array([], dtype=self.dtype).reshape(0, self.n_classes)
        if not self.lazy:
            self._load_all_data()

    def _load_all_data(self) -> NoReturn:
        """
        """
        # self.reader can not be pickled
        # with mp.Pool(processes=max(1, mp.cpu_count()-2)) as pool:
        #     self._signals, self._labels = \
        #         zip(*pool.starmap(_load_record, [(self.reader, r, self.config) for r in self.records]))

        # self._signals = np.array([]).reshape(0, len(self.config.leads), self.siglen)
        # self._labels = np.array([]).reshape(0, self.n_classes)

        fdr = FastDataReader(self.reader, self.records, self.config, self.ppm)

        # with tqdm(self.records, desc="Loading data", unit="records") as pbar:
        #     for rec in pbar:
                # s, l = self._load_one_record(rec)  # self._load_one_record is much slower than FastDataReader
        self._signals, self._labels = [], []
        with tqdm(desc="Loading data", total=len(fdr), unit="records") as pbar:
            for idx in range(len(fdr)):
                s, l = fdr[idx]
                # np.concatenate slows down the process severely
                # self._signals = np.concatenate((self._signals, s), axis=0)
                # self._labels = np.concatenate((self._labels, l), axis=0)
                self._signals.append(s)
                self._labels.append(l)
                pbar.update(1)
        self._signals = np.concatenate(self._signals, axis=0).astype(self.dtype)
        self._labels = np.concatenate(self._labels, axis=0)

    def _load_one_record(self, rec:str) -> Tuple[np.ndarray, np.ndarray]:
        """ finished, checked,

        load a record from the database using data reader

        NOTE
        ----
        DO NOT USE THIS FUNCTION DIRECTLY for preloading data,
        use `FastDataReader` instead

        Parameters
        ----------
        rec: str,
            the record to load

        Returns
        -------
        values: np.ndarray,
            the values of the record
        labels: np.ndarray,
            the labels of the record
        """
        values = self.reader.load_resampled_data(
            rec,
            leads=self.config.leads,
            # leads=Standard12Leads,
            data_format=self.config.data_format,
            siglen=None
        )
        for l in range(values.shape[0]):
            values[l] = remove_spikes_naive(values[l])
        values, _ = self.ppm(values, self.config.fs)
        values = ensure_siglen(
            values,
            siglen=self.siglen,
            fmt=self.config.data_format,
            tolerance=self.config.sig_slice_tol,
        ).astype(self.dtype)
        if values.ndim == 2:
            values = values[np.newaxis, ...]
        
        labels = self.reader.get_labels(
            rec, scored_only=True, fmt="a", normalize=True
        )
        labels = np.isin(self.all_classes, labels).astype(self.dtype)[np.newaxis, ...].repeat(values.shape[0], axis=0)

        return values, labels

    def to(self, leads:Sequence[str]) -> NoReturn:
        """
        """
        prev_leads = self.config.leads
        self.config.leads = leads
        self._indices = [prev_leads.index(l) for l in leads]
        self._signals = self._signals[:, self._indices, :]

    def emtpy(self, leads:Optional[Sequence[str]]=None) -> NoReturn:
        """
        """
        if leads is None:
            leads = self.config.leads
        else:
            self.config.leads = leads
        self._signals = np.array([], dtype=self.dtype).reshape(0, len(leads), self.siglen)

    @classmethod
    def from_extern(cls, ext_ds:"CINC2021", config:CFG) -> "CINC2021":
        """
        """
        new_ds = cls(config, ext_ds.training, lazy=True)
        indices = [ext_ds.config.leads.index(l) for l in new_ds.config.leads]
        new_ds._signals = ext_ds._signals[:, indices, :]
        new_ds._labels = ext_ds._labels.copy()
        return new_ds

    def reload_from_extern(self, ext_ds:"CINC2021") -> NoReturn:
        """
        """
        indices = [ext_ds.config.leads.index(l) for l in self.config.leads]
        self._signals = ext_ds._signals[:, indices, :]
        self._labels = ext_ds._labels.copy()

    @property
    def signals(self) -> np.ndarray:
        """
        """
        return self._signals

    @property
    def labels(self) -> np.ndarray:
        """
        """
        return self._labels

    def __getitem__(self, index:int) -> Tuple[np.ndarray, np.ndarray]:
        """ finished, checked,
        """
        return self.signals[index], self.labels[index]

    def __len__(self) -> int:
        """
        """
        return len(self._signals)
    
    def _train_test_split(self,
                          train_ratio:float=0.8,
                          force_recompute:bool=False) -> List[str]:
        """ finished, checked,

        do train test split,
        it is ensured that both the train and the test set contain all classes

        Parameters
        ----------
        train_ratio: float, default 0.8,
            ratio of the train set in the whole dataset (or the whole tranche(s))
        force_recompute: bool, default False,
            if True, force redo the train-test split,
            regardless of the existing ones stored in json files

        Returns
        -------
        records: list of str,
            list of the records split for training or validation
        """
        time.sleep(1)
        start = time.time()
        print("\nstart performing train test split...\n")
        time.sleep(1)
        _TRANCHES = list("ABEFG")
        _train_ratio = int(train_ratio*100)
        _test_ratio = 100 - _train_ratio
        assert _train_ratio * _test_ratio > 0

        ns = "_ns" if len(self.config.special_classes) == 0 else ""
        file_suffix = f"_siglen_{self.siglen}{ns}.json"
        train_file = self.reader.db_dir_base / f"train_ratio_{_train_ratio}{file_suffix}"
        test_file = self.reader.db_dir_base / f"test_ratio_{_test_ratio}{file_suffix}"

        # TODO: use self.reader.df_stats (precomputed and stored in utils/stats.csv)
        # to accelerate the validity examinations
        if force_recompute or not all([train_file.is_file(), test_file.is_file()]):
            tranche_records = {t: [] for t in _TRANCHES}
            train_set = {t: [] for t in _TRANCHES}
            test_set = {t: [] for t in _TRANCHES}
            for t in _TRANCHES:
                with tqdm(self.reader.all_records[t], total=len(self.reader.all_records[t])) as bar:
                    for rec in bar:
                        if rec in self.reader.exceptional_records:
                            # skip exceptional records
                            continue
                        rec_labels = self.reader.get_labels(
                            rec,
                            scored_only=True,
                            fmt="a",
                            normalize=True
                        )
                        rec_labels = [c for c in rec_labels if c in self.config.tranche_classes[t]]
                        if len(rec_labels) == 0:
                            # skip records with no scored class
                            continue
                        rec_samples = self.reader.load_resampled_data(rec).shape[1]
                        # NEW in CinC2021 compared to CinC2020
                        # training input siglen raised from 4000 to 5000,
                        # hence allow tolerance in siglen now
                        if rec_samples < self.siglen - self.config.input_len_tol:
                            continue
                        tranche_records[t].append(rec)
                time.sleep(1)
                print(f"tranche {t} has {len(tranche_records[t])} valid records for training")
            for t in _TRANCHES:
                is_valid = False
                while not is_valid:
                    shuffle(tranche_records[t])
                    split_idx = int(len(tranche_records[t])*train_ratio)
                    train_set[t] = tranche_records[t][:split_idx]
                    test_set[t] = tranche_records[t][split_idx:]
                    is_valid = self._check_train_test_split_validity(
                        train_set[t], test_set[t], set(self.config.tranche_classes[t])
                    )
            train_file.write_text(json.dumps(train_set, ensure_ascii=False))
            test_file.write_text(json.dumps(test_set, ensure_ascii=False))
            print(textwrap.dedent(f"""
                train set saved to \n\042{str(train_file)}\042
                test set saved to \n\042{str(test_file)}\042
                """
            ))
        else:
            train_set = json.loads(train_file.read_text())
            test_set = json.loads(test_file.read_text())

        print(f"train test split finished in {(time.time()-start)/60:.2f} minutes")

        _tranches = list(self.tranches or "ABEFG")
        if self.training == "all":
            records = list_sum([train_set[k] for k in _tranches]) + list_sum([test_set[k] for k in _tranches])
        elif self.training is True:
            records = list_sum([train_set[k] for k in _tranches])
        else:
            records = list_sum([test_set[k] for k in _tranches])
        return records


    def _check_train_test_split_validity(self,
                                         train_set:List[str],
                                         test_set:List[str],
                                         all_classes:Set[str]) -> bool:
        """ finished, checked,

        the train-test split is valid iff
        records in both `train_set` and `test` contain all classes in `all_classes`

        Parameters
        ----------
        train_set: list of str,
            list of the records in the train set
        test_set: list of str,
            list of the records in the test set
        all_classes: set of str,
            the set of all classes for training

        Returns
        -------
        is_valid: bool,
            the split is valid or not
        """
        train_classes = set(list_sum([self.reader.get_labels(rec, fmt="a") for rec in train_set]))
        train_classes.intersection_update(all_classes)
        test_classes = set(list_sum([self.reader.get_labels(rec, fmt="a") for rec in test_set]))
        test_classes.intersection_update(all_classes)
        is_valid = (len(all_classes) == len(train_classes) == len(test_classes))
        print(textwrap.dedent(f"""
            all_classes:     {all_classes}
            train_classes:   {train_classes}
            test_classes:    {test_classes}
            is_valid:        {is_valid}
            """
        ))
        return is_valid


    def persistence(self) -> NoReturn:
        """ finished, checked,

        make the dataset persistent w.r.t. the tranches and the ratios in `self.config`
        """
        _TRANCHES = "ABEFG"
        if self.training:
            ratio = int(self.config.train_ratio*100)
        else:
            ratio = 100 - int(self.config.train_ratio*100)
        fn_suffix = f"tranches_{self.tranches or _TRANCHES}_ratio_{ratio}"
        if self.config.bandpass is not None:
            bp_low = max(0, self.config.bandpass[0])
            bp_high = min(self.config.bandpass[1], self.config.fs//2)
            fn_suffix = fn_suffix + f"_bp_{bp_low:.1f}_{bp_high:.1f}"
        fn_suffix = fn_suffix + f"_siglen_{self.siglen}"

        X, y = [], []
        with tqdm(range(self.__len__()), total=self.__len__()) as bar:
            for idx in bar:
                values, labels = self.__getitem__(idx)
                X.append(values)
                y.append(labels)
        X, y = np.array(X), np.array(y)
        print(f"X.shape = {X.shape}, y.shape = {y.shape}")
        filename = f"{'train' if self.training else 'test'}_X_{fn_suffix}.npy"
        np.save(self.reader.db_dir_base / filename, X)
        print(f"X saved to {filename}")
        filename = f"{'train' if self.training else 'test'}_y_{fn_suffix}.npy"
        np.save(self.reader.db_dir_base / filename, y)
        print(f"y saved to {filename}")

    def _check_nan(self) -> NoReturn:
        """ finished, checked,

        during training, sometimes nan values are encountered,
        which ruins the whole training process
        """
        for idx, (values, labels) in enumerate(self):
            if np.isnan(values).any():
                print(f"values of {self.records[idx]} have nan values")
            if np.isnan(labels).any():
                print(f"labels of {self.records[idx]} have nan values")

    def extra_repr_keys(self) -> List[str]:
        return ["training", "tranches", "reader",]


class FastDataReader(ReprMixin, Dataset):
    """
    """
    def __init__(self, reader:CR, records:Sequence[str], config:CFG, ppm:Optional[PreprocManager]=None) -> NoReturn:
        """
        """
        self.reader = reader
        self.records = records
        self.config = config
        self.ppm = ppm
        if self.config.torch_dtype == torch.float64:
            self.dtype = np.float64
        else:
            self.dtype = np.float32

    def __len__(self) -> int:
        """
        """
        return len(self.records)

    def __getitem__(self, index:int) -> Tuple[np.ndarray, np.ndarray]:
        """
        """
        rec = self.records[index]
        values = self.reader.load_resampled_data(
            rec,
            leads=self.config.leads,
            # leads=Standard12Leads,
            data_format=self.config.data_format,
            siglen=None
        )
        for l in range(values.shape[0]):
            values[l] = remove_spikes_naive(values[l])
        if self.ppm:
            values, _ = self.ppm(values, self.config.fs)
        values = ensure_siglen(
            values,
            siglen=self.config.input_len,
            fmt=self.config.data_format,
            tolerance=self.config.sig_slice_tol,
        ).astype(self.dtype)
        if values.ndim == 2:
            values = values[np.newaxis, ...]
        
        labels = self.reader.get_labels(
            rec, scored_only=True, fmt="a", normalize=True
        )
        labels = np.isin(self.config.all_classes, labels).astype(self.dtype)[np.newaxis, ...].repeat(values.shape[0], axis=0)

        return values, labels

    def extra_repr_keys(self) -> List[str]:
        return ["reader", "ppm",]


def _load_record(reader:CR, rec:str, config:CFG) -> Tuple[np.ndarray, np.ndarray]:
    """ finished, NOT checked,

    load a record from the database using data reader

    Parameters
    ----------
    reader: CR,
        the data reader
    rec: str,
        the record to load
    config: dict,
        the configuration for loading record

    Returns
    -------
    values: np.ndarray,
        the values of the record
    labels: np.ndarray,
        the labels of the record
    """
    values = reader.load_resampled_data(
        rec,
        leads=config.leads,
        data_format="channel_first",
        siglen=None
    )
    values = ensure_siglen(values, siglen=config.input_len, fmt="channel_first")
    
    labels = reader.get_labels(
        rec, scored_only=True, fmt="a", normalize=True
    )
    labels = np.isin(config.all_classes, labels).astype(int)

    if config.data_format.lower() in ["channel_last", "lead_last"]:
        values = values.T

    return values, labels



# -*- coding: utf-8 -*-
"""
"""

from pathlib import Path
from collections import defaultdict
from typing import Union, Optional, Any, List, NoReturn, Dict, Sequence
from numbers import Real

import wfdb
import numpy as np
np.set_printoptions(precision=5, suppress=True)
import pandas as pd

from ..base import (
    PhysioNetDataBase,
    WFDB_Beat_Annotations, WFDB_Non_Beat_Annotations,
    BeatAnn,
)
from ...utils.misc import get_record_list_recursive3
from ...utils.utils_interval import generalized_intervals_intersection


__all__ = [
    "MITDB",
]


class MITDB(PhysioNetDataBase):
    """ NOT finished,

    MIT-BIH Arrhythmia Database

    ABOUT mitdb
    -----------
    1. contains 48 half-hour excerpts of two-channel ambulatory ECG recordings, obtained from 47 subjects.
    2. recordings were digitized at 360 samples per second per channel with 11-bit resolution over a 10 mV range.
    3. annotations contains:
        - beat-wise or finer (e.g. annotations of flutter wave) annotations, accessed via the `symbol` attribute of an `Annotation`.
        - rhythm annotations, accessed via the `aux_note` attribute of an `Annotation`.

    NOTE
    ----

    ISSUES
    ------

    Usage
    -----
    1.

    References
    ----------
    [1] https://physionet.org/content/mitdb/1.0.0/
    """
    def __init__(self,
                 db_dir:Optional[Union[str,Path]]=None,
                 working_dir:Optional[Union[str,Path]]=None,
                 verbose:int=2,
                 **kwargs:Any) -> NoReturn:
        """

        Parameters
        ----------
        db_dir: str or Path, optional,
            storage path of the database
            if not specified, data will be fetched from Physionet
        working_dir: str or Path, optional,
            working directory, to store intermediate files and log file
        verbose: int, default 2,
            log verbosity
        kwargs: auxilliary key word arguments
        """
        super().__init__(db_name="mitdb", db_dir=db_dir, working_dir=working_dir, verbose=verbose, **kwargs)
        self.fs = 360
        self.data_ext = "dat"
        self.ann_ext = "atr"

        self.beat_types_extended = list("""!"+/AEFJLNQRSV[]aefjx|~""")
        self.nonbeat_types = [
            item for item in self.beat_types_extended if item in WFDB_Non_Beat_Annotations
        ]
        self.beat_types = [
            item for item in self.beat_types_extended if item in WFDB_Beat_Annotations
        ]
        self.beat_types_map = {
            item: i for i, item in enumerate(self.beat_types)
        }
        self.beat_types_extended_map = {
            item: i for i, item in enumerate(self.beat_types_extended)
        }
        self.rhythm_types = [
            "(AB", "(AFIB", "(AFL", "(B", "(BII", "(IVR",
            "(N", "(NOD", "(P", "(PREX", "(SBR", "(SVTA",
            "(T", "(VFL", "(VT", "MISSB", "PSE", "TS",
        ]
        self.rhythm_types_map = {
            item.lstrip("("): i for i, item in enumerate(self.rhythm_types) \
                if item.startswith("(")
        }
        self._rhythm_ignore_index = -100

        self.all_leads = ["MLII", "V1", "V2", "V4", "V5"]

        self._ls_rec()

    def _ls_rec(self) -> NoReturn:
        """
        """
        super()._ls_rec()
        if len(self._all_records) == 0:
            self._all_records = get_record_list_recursive3(self.db_dir, f"^[\d]{{3}}.{self.data_ext}$")

    def load_data(self,
                  rec:str,
                  leads:Optional[Union[str, List[str]]]=None,
                  sampfrom:Optional[int]=None,
                  sampto:Optional[int]=None,
                  data_format:str="channel_first",
                  units:str="mV",
                  fs:Optional[Real]=None,) -> np.ndarray:
        """ finished, checked,

        load physical (converted from digital) ecg data,
        which is more understandable for humans

        Parameters
        ----------
        rec: str,
            name of the record
        leads: str or list of str, optional,
            the leads to load
        sampfrom: int, optional,
            start index of the data to be loaded
        sampto: int, optional,
            end index of the data to be loaded
        data_format: str, default "channel_first",
            format of the ecg data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first")
        units: str, default "mV",
            units of the output signal, can also be "μV", with an alias of "uV"
        fs: real number, optional,
            if not None, the loaded data will be resampled to this frequency
        
        Returns
        -------
        data: ndarray,
            the ecg data
        """
        fp = str(self.db_dir / rec)
        if not leads:
            _leads = self._get_lead_names(rec)
        elif isinstance(leads, str):
            _leads = [leads]
        else:
            _leads = leads
        assert set(_leads).issubset(self.all_leads)
        # p_signal in the format of "lead_last", and in units "mV"
        data = wfdb.rdrecord(
            fp,
            sampfrom=sampfrom or 0,
            sampto=sampto,
            physical=True,
            channel_names=_leads,
        ).p_signal
        if units.lower() in ["μv", "uv"]:
            data = 1000 * data
        if fs is not None and fs != self.fs:
            data = resample_poly(data, fs, self.fs, axis=0)
        if data_format.lower() in ["channel_first", "lead_first"]:
            data = data.T
        return data

    def load_ann(self,
                 rec:str,
                 sampfrom:Optional[int]=None,
                 sampto:Optional[int]=None,
                 rhythm_format:str="interval",
                 beat_format:str="beat",
                 extended_beats:bool=False,
                 keep_original:bool=False,) -> dict:
        """ finished, checked,

        load rhythm and beat annotations,
        which are stored in the `aux_note`, `symbol` attributes of corresponding annotation files.
        NOTE that qrs annotations (.qrs files) do NOT contain any rhythm annotations
        
        Parameters
        ----------
        rec: str,
            name of the record
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        rhythm_format: str, default "interval", case insensitive,
            format of returned annotation, can also be "mask"
        beat_format: str, default "beat", case insensitive,
            format of returned annotation, can also be "dict"
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        
        Returns
        -------
        ann, dict,
            the annotations of `rhythm` and `beat`, with
            `rhythm` annotatoins in the format of `intervals`, or `mask`;
            `beat` annotations in the format of `dict` or `BeatAnn`
        """
        assert rhythm_format.lower() in ["interval", "mask",], \
            f"`rhythm_format` must be one of ['interval', 'mask'], got {rhythm_format}"
        assert beat_format.lower() in ["beat", "dict",], \
            f"`beat_format` must be one of ['beat', 'dict'], got {beat_format}"
        fp = str(self.db_dir / rec)
        wfdb_ann = wfdb.rdann(fp, extension=self.ann_ext)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        sample_inds = wfdb_ann.sample
        indices = np.where((sample_inds >= sf) & (sample_inds < st))[0]

        beat_ann = [
            BeatAnn(i, s) for i, s in \
                zip(sample_inds[indices], np.array(wfdb_ann.symbol)[indices])
        ]

        rhythm_intervals = defaultdict(list)
        start_idx, rhythm = None, None
        for ra, si in zip(wfdb_ann.aux_note, sample_inds):
            ra = ra.rstrip("\x00")
            if ra.startswith("("):
                if start_idx is not None:
                    rhythm_intervals[rhythm].append([start_idx, si])
                start_idx = si
                rhythm = ra.lstrip("(")
        if start_idx is not None:
            rhythm_intervals[rhythm].append([start_idx, si])
            # rhythm_intervals[rhythm].append([start_idx, sig_len])
        rhythm_intervals = {
            k: np.array(generalized_intervals_intersection(v, [[sf, st]])) \
                for k, v in rhythm_intervals.items()
        }
        if rhythm_format.lower() == "mask":
            rhythm_mask = np.full((st-sf,), self._rhythm_ignore_index, dtype=int)
            for k, v in rhythm_intervals.items():
                for itv in v:
                    rhythm_mask[itv[0]-sf:itv[1]-sf] = self.rhythm_types_map[k]

        if not keep_original:
            rhythm_intervals = {
                k: v - sf for k, v in rhythm_intervals.items()
            }
            for b in beat_ann:
                b.index -= sf

        if not extended_beats:
            beat_ann = [
                b for b in beat_ann if b.symbol in self.beat_types
            ]

        if beat_format.lower() == "dict":
            beat_ann = {
                s: np.array([b.index for b in beat_ann if b.symbol==s], dtype=int) \
                    for s in self.beat_types_extended
            }
            beat_ann = {k: v for k, v in beat_ann.items() if len(v) > 0}

        ann = {}
        ann["beat"] = beat_ann
        if rhythm_format.lower() == "interval":
            ann["rhythm"] = rhythm_intervals
        else:
            ann["rhythm"] = rhythm_mask
        return ann

    def load_rhythm_ann(self,
                        rec:str,
                        sampfrom:Optional[int]=None,
                        sampto:Optional[int]=None,
                        rhythm_format:str="interval",
                        keep_original:bool=False,) -> Union[Dict[str, list], np.ndarray]:
        """  finished, checked,

        load rhythm annotations,
        which are stored in the `aux_note` attribute of corresponding annotation files.
        
        Parameters
        ----------
        rec: str,
            name of the record
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        rhythm_format: str, default "interval", case insensitive,
            format of returned annotation, can also be "mask"
        beat_format: str, default "beat", case insensitive,
            format of returned annotation, can also be "dict"
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        
        Returns
        -------
        ann, dict or ndarray,
            the annotations in the format of intervals, or in the format of mask
        """
        return self.load_ann(
            rec, sampfrom, sampto,
            rhythm_format=rhythm_format,
            keep_original=keep_original,
        )["rhythm"]

    def load_beat_ann(self,
                      rec:str,
                      sampfrom:Optional[int]=None,
                      sampto:Optional[int]=None,
                      beat_format:str="beat",
                      keep_original:bool=False,) -> Union[Dict[str, np.ndarray], List[BeatAnn]]:
        """ finished, checked,

        load beat annotations,
        which are stored in the `symbol` attribute of corresponding annotation files
        
        Parameters
        ----------
        rec: str,
            name of the record
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        beat_format: str, default "beat", case insensitive,
            format of returned annotation, can also be "dict"
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        
        Returns
        -------
        ann, dict or list,
            locations (indices) of the all the beat types ("A", "N", "Q", "V",)
        """
        return self.load_ann(
            rec, sampfrom, sampto,
            beat_format=beat_format,
            keep_original=keep_original,
        )["beat"]

    def load_rpeak_indices(self,
                           rec:str,
                           sampfrom:Optional[int]=None,
                           sampto:Optional[int]=None,
                           keep_original:bool=False,) -> np.ndarray:
        """ finished, checked,

        load rpeak indices, or equivalently qrs complex locations,
        which are stored in the `symbol` attribute of corresponding annotation files,
        regardless of their beat types,
        
        Parameters
        ----------
        rec: str,
            name of the record
        sampfrom: int, optional,
            start index of the annotations to be loaded
        sampto: int, optional,
            end index of the annotations to be loaded
        keep_original: bool, default False,
            if True, indices will keep the same with the annotation file
            otherwise subtract `sampfrom` if specified
        
        Returns
        -------
        rpeak_inds, ndarray,
            locations (indices) of the all the rpeaks (qrs complexes)
        """
        fp = str(self.db_dir / rec)
        wfdb_ann = wfdb.rdann(fp, extension=self.ann_ext)
        header = wfdb.rdheader(fp)
        sig_len = header.sig_len
        sf = sampfrom or 0
        st = sampto or sig_len
        assert st > sf, "`sampto` should be greater than `sampfrom`!"

        rpeak_inds = wfdb_ann.sample
        indices = np.where(
            (rpeak_inds >= sf) & (rpeak_inds < st) \
                & (np.isin(wfdb_ann.symbol, self.beat_types))
        )[0]
        rpeak_inds = rpeak_inds[indices]
        if not keep_original:
            rpeak_inds -= sf
        return rpeak_inds

    def _get_lead_names(self, rec:str) -> List[str]:
        """

        Parameters
        ----------
        rec: str,
            name of the record
        
        returns:
        list of str:
            a list of names of the leads contained in the record
        """
        fp = str(self.db_dir / rec)
        return wfdb.rdheader(fp).sig_name

    def plot(self,
             rec:str,
             data:Optional[np.ndarray]=None,
             ann:Optional[Dict[str, np.ndarray]]=None,
             beat_ann:Optional[Dict[str, np.ndarray]]=None,
             rpeak_inds:Optional[Union[Sequence[int],np.ndarray]]=None,
             ticks_granularity:int=0,
             leads:Optional[Union[int, List[int]]]=None,
             sampfrom:Optional[int]=None,
             sampto:Optional[int]=None,
             same_range:bool=False,
             **kwargs:Any) -> NoReturn:
        """
        """
        raise NotImplementedError



"""
Higher Order ResNet

References
----------
[1] Luo, Z., Sun, Z., Zhou, W., & Kamata, S. I. (2021). Rethinking ResNets: Improved Stacking Strategies With High Order Schemes. arXiv preprint arXiv:2103.15244.
"""

from copy import deepcopy
from itertools import repeat
from typing import Union, Optional, Sequence, NoReturn

import numpy as np
np.set_printoptions(precision=5, suppress=True)
import torch
from torch import nn
from torch import Tensor

from ...cfg import CFG, DEFAULTS
from ...utils.utils_nn import compute_module_size, SizeMixin
from ...utils.misc import dict_to_str
from ...models._nets import (
    Activations,
    Conv_Bn_Activation,
    DownSample,
    ZeroPadding,
    NonLocalBlock, SEBlock, GlobalContextBlock,
)

if DEFAULTS.torch_dtype == torch.float64:
    torch.set_default_tensor_type(torch.DoubleTensor)


__all__ = [
    "MidPointResNet",
    "RK4ResNet",
    "RK8ResNet",
]


class MidPointResNet(SizeMixin, nn.Module):
    """
    """
    def __init__(self, in_channels:int, **config) -> NoReturn:
        """
        """
        raise NotImplementedError


class RK4ResNet(SizeMixin, nn.Module):
    """
    """
    def __init__(self, in_channels:int, **config) -> NoReturn:
        """
        """
        raise NotImplementedError


class RK8ResNet(SizeMixin, nn.Module):
    """
    """
    def __init__(self, in_channels:int, **config) -> NoReturn:
        """
        """
        raise NotImplementedError



"""
the modules that follows CNN feature extractor,
mainly RNN, but can also be attention
"""

from copy import deepcopy

from ..cfg import CFG


__all__ = [
    "lstm",
    "attention",
]


lstm = CFG()
lstm.bias = True
lstm.dropouts = 0.2
lstm.bidirectional = True
lstm.retseq = True
lstm.hidden_sizes = [12*16, 12*16]


attention = CFG()
# almost the same with lstm, but the last layer is an attention layer
attention.head_num = 12
attention.bias = True
attention.dropouts = 0.2
attention.bidirectional = True
attention.hidden_sizes = [12*24, 12*6]
